(ns puppetlabs.redrover.import
  (:require [cheshire.core :as json]
            [clj-yaml.core :as yaml]
            [clojure.java.shell :as shell]
            [clojure.pprint :refer [pprint]]
            [clojure.string :as str]
            [clojure.tools.logging :as log]
            [puppetlabs.certificate-authority.core :as ssl]
            [puppetlabs.classifier.http :refer [convert-uuids]]
            [puppetlabs.classifier.storage :as storage]
            [puppetlabs.http.client.sync :as http]
            [puppetlabs.redrover.file :refer [read-yaml-file]]
            [puppetlabs.redrover.group :as group]
            [puppetlabs.redrover.schema :refer [ConvertedHierarchy DashboardExport]]
            [schema.core :as sc]
            [slingshot.slingshot :refer [throw+ try+]])
  (:import com.fasterxml.jackson.core.JsonParseException))

;; API configuration

(def default-terminus-config-path "/etc/puppetlabs/puppet/classifier.yaml")

(defn terminus-config->api-url
  "Given a map representing classifier terminus configuration return the
  base url for the classifier API."
  [{:keys [server port prefix]}]
  (str "https://" server ":" port prefix))

(defn terminus-config-file->api-url
  "Read the classifier terminus configuration file and return the base
  url for the classifier API."
  [path]
  (terminus-config->api-url (read-yaml-file path)))


;; SSL Configuration

(def default-puppet-binary "/opt/puppet/bin/puppet")

(defn run-puppet
  [puppet-bin & args]
  (try+ (apply shell/sh (cons puppet-bin args))
    (catch java.io.IOException e
      (throw+ {:kind ::shell-error
               :msg (str "Unable to run puppet due to the following error: "
                         (str/lower-case (.getMessage e)))}))))

(defn get-puppet-setting!
  "Retrieve one setting from puppet"
  [puppet-bin setting]
  (let [return (run-puppet puppet-bin "config" "print" setting)]
    (when (not (= 0 (:exit return)))
      (throw+ {:kind ::puppet-error
               :msg (str "Got exit code " (:exit return)
                         " trying to run \"" puppet-bin
                         "\" config print " setting ".")}))
    (-> return
      :out
      str/trim-newline)))

(defn get-puppet-ssl-paths!
  "Given a path to a puppet binary, return the paths it uses for ssl
  configuration."
  [puppet-bin]
  {:ssl-key (get-puppet-setting! puppet-bin "hostprivkey")
   :ssl-cert (get-puppet-setting! puppet-bin "hostcert")
   :ssl-ca-cert (get-puppet-setting! puppet-bin "cacert")})

(defn init-ssl-context-from-puppet-pems
  [{:keys [ssl-cert ssl-key ssl-ca-cert]}]
  (try+ (ssl/pems->ssl-context ssl-cert ssl-key ssl-ca-cert)
    (catch java.io.IOException e
      (throw+ {:kind ::unreadable-certificate
               :msg (str "Unable to read one or more PEM files "
                         "due to the following error: "
                         (str/lower-case (.getMessage e)) ". "
                         "Please check the --ssl-key, --ssl-path, and "
                         "--ssl-ca-cert settings.")}))))

(defn merge-ssl-defaults
  "Merge default options for ssl context into the options map"
  [{:keys [ssl-key ssl-cert ssl-ca-cert] :as options}]
  (let [puppet-bin (or (:puppet options) default-puppet-binary)
        puppet-ssl-options (if (some nil? [ssl-key ssl-cert ssl-ca-cert])
                             (get-puppet-ssl-paths! puppet-bin))]
    (merge puppet-ssl-options options)))

(defn merge-api-url-default
  [{:keys [classifier-api-url] :as options}]
  (if (nil? classifier-api-url)
    (assoc options
           :classifier-api-url
           (terminus-config-file->api-url default-terminus-config-path))
    options))

(defn merge-ssl-context
  [options]
  (assoc options
         :ssl-context
         (init-ssl-context-from-puppet-pems options)))

(defn merge-options-with-defaults
  [options]
  (-> options
      merge-ssl-defaults
      merge-ssl-context
      merge-api-url-default))

(defn- http-fn
  "Returns a closure that wraps the given http function, throwing an
  ::http-exception with an informative message if there's an IOException.

  The closure assumes that the wrapped function has two arities:
  [url] and [url options]."
  [f]
  (fn [url & options]
    (try+ (apply f url options)
      (catch java.io.IOException e
        (log/error (.getMessage e))
        (throw+ {:kind ::http-exception
                 :msg (str "Unable to contact the classifier at: " url ". "
                           "Please check the --classifier-api-url setting and "
                           "make sure that the service is running and accessible.")})))))

(def ^{:arglists '([url] [url opts])} http-get (http-fn http/get))

(def ^{:arglists '([url] [url opts])} http-post (http-fn http/post))

;; Importing

(defn- parse-body
  "This function takes an HTTP response and attempts to parse the body as JSON.
  If the JSON parsing was unsuccessful, just returns the body as a string."
  [resp]
  (let [{body :body} resp
        body-str (if (isa? (class body) java.io.InputStream)
                   (slurp body)
                   (str body))]
    (try (json/decode body-str true)
      (catch JsonParseException _
        body-str))))

(defn get-hierarchy
  [classifier-api-url ssl-context]
  (let [groups-url (str classifier-api-url "/v1/groups")
        {:keys [status] :as response} (http-get groups-url
                                                {:ssl-context ssl-context
                                                 :headers {"Accept" "application/json"}})
        body (parse-body response)]

    (when-not (= 200 status)
      (throw+ {:kind ::unexpected-response
               :msg (str "Received unexpected response from ", groups-url
                         ": HTTP Code ", status, ".\n"
                         (if (string? body)
                           body
                           (with-out-str (pprint body))))}))

    (when (string? body)
      (throw+ {:kind ::unexpected-response
               :msg (str "Received a 200 response from " groups-url " that was not JSON."
                         " The classifier API URL is misconfigured."
                         "\n" body)}))

    body))

(def agent-specified-env-group-name "Agent-specified environment")
(def prod-env-group-name "Production environment")

(def infrastructure-group-names
  #{"default"
    agent-specified-env-group-name
    prod-env-group-name
    "PE ActiveMQ Broker"
    "PE Certificate Authority"
    "PE Console"
    "PE Infrastructure"
    "PE Master"
    "PE MCollective"
    "PE PuppetDB"
    "PE Agent"})

(defn agent-specified-environment-group?
  [group]
  (and (= true (:environment-trumps group))
       (= "agent-specified" (:environment group))))

(defn merge-hierarchy
  [existing-groups transformed-groups]
  (let [infra-groups (filter #(infrastructure-group-names (:name %)) existing-groups)
        [prod-group] (filter #(= prod-env-group-name (:name %)) infra-groups)
        prod-group (if prod-group
                     (assoc prod-group :rule ["~" "name" ".*"])
                     (assoc (group/blank-env-group prod-env-group-name)
                            :environment "production"
                            :description "By default, nodes use the production environment."
                            :rule ["~" "name" ".*"]))
        [ag-spec-group] (filter #(= agent-specified-env-group-name (:name %)) infra-groups)
        ag-spec-group (if ag-spec-group
                        (dissoc ag-spec-group :rule)
                        (assoc (group/blank-env-group agent-specified-env-group-name)
                               :description "Nodes in this group set their own environment."
                               :parent (:id prod-group)))
        infra-groups' (concat [prod-group ag-spec-group]
                              (->> infra-groups
                                (remove #(= prod-env-group-name (:name %)))
                                (remove #(= agent-specified-env-group-name (:name %)))))
        transformed-groups' (->> transformed-groups
                              (remove storage/root-group?)
                              (map #(if (agent-specified-environment-group? %)
                                      (assoc % :parent (:id prod-group))
                                      %)))]
    (concat infra-groups' transformed-groups')))

(defn post-hierarchy
  [classifier-api-url ssl-context hierarchy-groups]
  (let [import-url (str classifier-api-url "/v1/import-hierarchy")
        {:keys [status] :as response} (http-post import-url
                                                 {:headers {"Content-Type" "application/json"}
                                                  :body (json/encode hierarchy-groups)
                                                  :ssl-context ssl-context})
        body (parse-body response)]
    (when-not (= 204 status)
      (throw+ {:kind ::unexpected-response
               :msg (str "Received unexpected response from ", import-url
                         ": HTTP Code ", status, ".\n",
                         (if (string? body)
                           body
                           (with-out-str (pprint body))))}))))

(defn validate-groups-and-convert-uuids
  [transformed-groups]
  (let [transformed-groups-with-uuids (map convert-uuids transformed-groups)]
    (when-let [validation-failures (sc/check ConvertedHierarchy transformed-groups-with-uuids)]
      (log/debug (str "The input to the import command caused the following schema failure: "
                      validation-failures))

      (if (nil? (sc/check DashboardExport transformed-groups))
        (throw+ {:kind ::invalid-input-file
                 :msg (str "The input file must be converted before import.\n"
                           "Please use the \"convert\" command to prepare this"
                           " classification data first and then run the \"import\""
                           " command again.")})

        (throw+ {:kind ::invalid-input-file
                 :msg (str "The input file contains valid JSON, but it does not appear"
                           " to contain converted classification data.\n"
                           "Please re-run this step using the output of the \"convert\" command"
                           " as the input.")})))
    transformed-groups-with-uuids))

(defn import!
  [{:keys [classifier-api-url ssl-context]} transformed-groups]
  (let [transformed-groups-with-uuids (validate-groups-and-convert-uuids transformed-groups)
        existing-groups (get-hierarchy classifier-api-url ssl-context)
        new-groups (merge-hierarchy existing-groups transformed-groups-with-uuids)]
    (post-hierarchy classifier-api-url ssl-context new-groups)))
